--[[===========================================================================
ChordMemory.lua
===========================================================================]]--

return {
arguments = {
  {
      name = "mode",
      value = 2,
      properties = {
          max = 2,
          min = 1,
          display_as = "switch",
          items = {
              "read",
              "write",
          },
      },
      description = "Set the model operating mode \n'read' = interpret existing pattern data \n'write' = expand existing notes into chords",
  },
  {
      name = "num_keys",
      value = 3,
      properties = {
          max = 6,
          min = 1,
          display_as = "switch",
          items = {
              "2",
              "3",
              "4",
              "5",
              "6",
              "7",
          },
      },
      description = "Number of keys in this chord",
  },
  {
      name = "duration",
      value = 1.5,
      properties = {
          max = 64,
          display_as = "float",
          min = 0,
      },
      description = "Duration of notes (0 = infinite)",
  },
  {
      name = "interval_2",
      value = 2,
      properties = {
          max = 119,
          display_as = "integer",
          min = 0,
      },
      description = "Amount of semitones between 1 & 2",
  },
  {
      name = "interval_3",
      value = 3,
      properties = {
          max = 119,
          display_as = "integer",
          min = 0,
      },
      description = "Amount of semitones between 2 & 3",
  },
  {
      name = "interval_4",
      value = 3,
      properties = {
          max = 119,
          display_as = "integer",
          min = 0,
      },
      description = "Amount of semitones between 3 & 4",
  },
  {
      name = "interval_5",
      value = 2,
      properties = {
          max = 119,
          display_as = "integer",
          min = 0,
      },
      description = "Amount of semitones between 4 & 5",
  },
  {
      name = "interval_6",
      value = 2,
      properties = {
          max = 119,
          min = 0,
          display_as = "integer",
          zero_based = false,
      },
      description = "Amount of semitones between 5 & 6",
  },
  {
      name = "interval_7",
      value = 0,
      properties = {
          max = 119,
          min = 0,
          display_as = "integer",
          zero_based = false,
      },
      description = "Amount of semitones between 6 & 7",
  },
  {
      name = "strumming",
      value = 2,
      properties = {
          max = 3,
          min = 1,
          display_as = "switch",
          items = {
              "off",
              "up",
              "down",
          },
      },
      description = "Decide if/how notes are spread across lines",
  },
  {
      name = "strm_space",
      value = 1,
      properties = {
          max = 8,
          min = 1,
          display_as = "integer",
          zero_based = false,
      },
      description = "Space (in lines) between each strummed note",
  },
  {
      name = "human_vol",
      value = 49.315068493151,
      properties = {
          max = 100,
          display_as = "percent",
          min = 0,
      },
      description = "Humanize volume (relative to source note)",
  },
  {
      name = "human_pan",
      value = 54.794520547945,
      properties = {
          max = 100,
          display_as = "percent",
          min = 0,
      },
      description = "Humanize panning (relative to source note)",
  },
  {
      name = "human_dly",
      value = 25.753424657534,
      properties = {
          max = 100,
          display_as = "percent",
          min = 0,
      },
      description = "Humanize delay (relative to source note)",
  },
},
presets = {
  {
      num_keys = 3,
      interval_5 = 2,
      interval_2 = 2,
      interval_4 = 3,
      duration = 1.5,
      human_pan = 71.232876712329,
      human_dly = 25.753424657534,
      interval_3 = 3,
      strm_space = 1,
      name = "Strum test",
      interval_6 = 2,
      human_vol = 59.835616438356,
      mode = 2,
      strumming = 2,
      interval_7 = 0,
  },
},
data = {
},
options = {
 color = 0x000000,
},
callback = [[
-------------------------------------------------------------------------------
-- Chord Memory
-- (this version contains detection of note duration, might be handy...)
-------------------------------------------------------------------------------
if (args.mode == 1) then 

  -- initialize data
  if not data.recent_col_idx then
    print("initialize data")
    data.notes = {} -- note pitches
    data.lines = {} -- note xinc-value
    data.columns = {} -- note column indices
    data.note_duration = 0 -- default to 'endless'
  end

  local least_vol,most_vol = 0x80,0x00
  local least_pan,most_pan = 0x80,0x00
  local least_dly,most_dly = 0xFF,0x00
  
  print("xinc",xinc)
  
  for col_idx = 1,12 do
    local xcol = xline.note_columns[col_idx]
    if xcol then 
    
      -- determine note-on
    
      --print("col_idx",col_idx)
      if (xcol.note_value < 119) then
      
        -- save the progress through columns (this lets us resume
        -- search at a later line, when looking for strummed chords)
        data.recent_col_idx = col_idx
      
        table.insert(data.notes,xcol.note_value)
        table.insert(data.lines,xinc)
        table.insert(data.columns,col_idx)
        
        --print(">>> columns",rprint(data.columns))
        
        -- volume
        local vol = (xcol.volume_value <= 0x80) 
          and xcol.volume_value or 0x80
        least_vol = math.min(least_vol,vol)
        most_vol = math.max(most_vol,vol)
        
        -- panning
        local pan = (xcol.panning_value <= 0x80) 
          and xcol.panning_value or 0x40
        least_pan = math.min(least_pan,pan)
        most_pan = math.max(most_pan,pan)
        
        -- delay    
        local dly = xcol.delay_value 
        least_dly = math.min(least_dly,dly)
        most_dly = math.max(most_dly,dly)
        
      end
      
      -- determine note-off (duration)
      
      local previous_col = data.columns and table.find(data.columns,col_idx) 
      if previous_col and (data.note_duration == 0) then
      
        --print("volume_string",string.sub(xcol.volume_string,1,1))
        --print("panning_string",string.sub(xcol.panning_string,1,1))
      
        if (xcol.note_value == 120) then           
          -- note-off
          data.note_duration = (xinc - data.lines[previous_col]) 
            + (xcol.delay_value/0x100)
        elseif (string.sub(xcol.volume_string,1,1) == "C")
            or (string.sub(xcol.panning_string,1,1) == "C")
        then
          -- note-cut      
          print("found note-cut for previously triggered note")
          local volume_cut = (string.sub(xcol.volume_string,1,1) == "C")
            and tonumber(("0x%s"):format(string.sub(xcol.volume_string,2,2)))
          local panning_cut = (string.sub(xcol.panning_string,1,1) == "C")
            and tonumber(("0x%s"):format(string.sub(xcol.panning_string,2,2)))
          --print("volume_cut,panning_cut",volume_cut,panning_cut)
          local cut_at_tick = volume_cut or panning_cut
          data.note_duration = (xinc - data.lines[previous_col]) 
            + (cut_at_tick/rns.transport.tpl)          
        end
        
        if (data.note_duration > 0) then
          args.duration = tonumber(("%.4f"):format(data.note_duration))        
        end
                            
      end
      
    end
  end
  
  -- determine if we are strumming, and if so, spacing/direction...
  local line_spacing = 0
  local going_upwards = false
  if (#data.lines > 1) then
    for k = 2,#data.lines do
      line_spacing = math.max(line_spacing,data.lines[k]-data.lines[k-1])
      going_upwards = (data.notes[k] > data.notes[k-1])
    end
  end
  print("line_spacing",line_spacing)
  
  -- update arguments as information is received...
  
  -- sort notes before determining intervals    
  local sorted = table.copy(data.notes)
  table.sort(sorted) -- sort low -> high
  
  if (#sorted > 1) then
    args.num_keys = #sorted
    for k = 2,#sorted do
      args["interval_"..k] = sorted[k] - sorted[k-1] 
    end
    for k = #sorted+1,7 do -- set others to '0'
      args["interval_"..k] = 0
    end
    args.strumming = (line_spacing > 0) and (going_upwards and 2 or 3) or 1
    args.strm_space = (line_spacing > 1) and line_spacing or 1
    args.human_vol = ((most_vol-least_vol)/0x80)*100
    args.human_pan = ((most_pan-least_pan)/0x80)*100
    args.human_dly = ((most_dly-least_dly)/0xFF)*100
  end
  
else 

  -- produce output

  local xcol = xline.note_columns[1]
  if (xcol.note_value < 119) then
    data.source_xcol = xcol
    data.source_xinc = xinc
  end
  
  if data.source_xcol then
  
    local get_note_pitch = function(idx)
      local pitch = data.source_xcol.note_value
      for k = 1,idx-1 do
        pitch = pitch + args["interval_"..k+1] 
      end
      return pitch
    end

    local humanize = function(val,var,min,max)  
      print("humanize - val,var,min,max",val,var,min,max)    
      local range = max-min
      local random = math.random(0,range) * (var/100)
      val = math.floor(math.min(max,math.min(max,val+(random-(var/2)))))
      return val
    end
    
    local output_note = function(col_idx)

      xcol = xline.note_columns[col_idx]
      
      -- note/instrument
      xcol.note_value = get_note_pitch(col_idx)
      xcol.instrument_value = data.source_xcol.instrument_value
      
      -- volume (humanized relative to itself)
      local vol = (data.source_xcol.volume_value <= 0x80) 
        and data.source_xcol.volume_value or 0x80      
      local vol_var = args.human_vol * (vol/0x80)
      vol = humanize(vol,vol_var,0x00,0x80)
      xcol.volume_value = (vol == 0x80) and 255 or vol
      
      -- panning 
      local pan = (data.source_xcol.panning_value <= 0x80) 
        and data.source_xcol.panning_value or 0x40      
      pan = humanize(pan,args.human_pan,0x00,0x80)
      xcol.panning_value = (pan == 0x40) and 255 or pan
      
      -- delay
      local dly = data.source_xcol.delay_value      
      xcol.delay_value = humanize(dly,args.human_dly,0x00,0xFF)
      
    end    
    
    for col_idx = 2,args.num_keys+1 do
      local note_written = false
      if (args.strumming == 1) then
        if (data.source_xinc == xinc) then            
          output_note(col_idx)
          note_written = true
        end
      else -- strum up/down
        local strum_range = (args.num_keys*args.strm_space)
        local strum_to = data.source_xinc+strum_range
        local count = 1
        for k = data.source_xinc,strum_to,args.strm_space do
          if (count == col_idx) and (xinc == k) then
            output_note(col_idx)
            note_written = true
          end
          count = count + 1
        end
      end
      
      if not note_written then
        xline.note_columns[col_idx] = {}
      end   
         
    end -- iterate through columns

    if (xinc == data.source_xinc) then
      for col_idx = args.num_keys+2,6 do
        local xcol = xline.note_columns[col_idx]
        if xcol then
          xline.note_columns[col_idx].note_string  = "OFF"
        end
      end
    end
        
  end
  
end








]],
}